import __future__
from collections import defaultdict
import numpy as np
import sys

PE_NUM = 4

class BlockParser:
  def __init__(self):
    self.status = ""
    self.last_time = 0

  def processBlock(self, block_name, time):
    # something like (processq2_init_b, processq2_init_e)
    block_body = block_name[:-2]
    assert(block_body)
    isBegin = (block_name[-1] == 'b')
    isEnd = (block_name[-1] == 'e')
    assert(isBegin ^ isEnd)

    elapsed_time = time - self.last_time
    last_block_body = self.status
    if last_block_body:
      assert(last_block_body == block_body)
      assert(isEnd)
      self.status = ""
      self.last_time = time
      return 1, block_body, elapsed_time
    else:
      assert(isBegin)
      self.status = block_body
      self.last_time = time
      return 0, block_body, elapsed_time

class LineParser:
  def __init__(self):
    self.status_fifo = list()
    self.last_time = 0

  def processTimeLine(self, line):
    assert(len(self.status_fifo)>0)
    block_name = self.status_fifo.pop(0)
    time = int(line.split('=')[1])
    assert(self.last_time < time)
    self.last_time = time
    return block_name, time

  def processBlockLine(self, line):
      block_name = line.split(':')[1]
      self.status_fifo.append(block_name)

  def size(self):
      return len(self.status_fifo)

class LineParserGroup:
  def __init__(self):
    self.PE_LineParser = list()
    self.PE_BlockParsers = list()

    for i in range(PE_NUM):
      self.PE_LineParser.append(LineParser())
      self.PE_BlockParsers.append(BlockParser())

  def get(self, i):
    assert(i < PE_NUM)
    return self.PE_LineParser[i]

  def getMinSize(self):
    minsize = 999999;
    for i in range(PE_NUM):
      minsize = min(minsize, self.get(i).size())
    return minsize

  def getFifo(self, i):
    return self.get(i).status_fifo

  def processBlockLine(self, i, line):
    return self.get(i).processBlockLine(line)

  def processTimeLine(self, i, line):
    block_name, time = self.get(i).processTimeLine(line)
    return block_name, time

  def processBlock(self, i, block_name, time):
    assert(i < PE_NUM)
    isBlock, block, elapsed_time = self.PE_BlockParsers[i].processBlock(block_name, time)
    return block_name, elapsed_time


class StatisticManager:
  def __init__(self):
    self.table = [""] * PE_NUM
    for i in range(PE_NUM):
        self.table[i] = defaultdict(list)
        
  def add(self, pe_id, block_name, elapsed_time):
    self.table[pe_id][block_name].append(elapsed_time)

  def hasBlock(self, pe_id, block_name):
    return (block_name in self.table[pe_id])

  def getArray(self, pe_id, block_name):
    assert(self.hasBlock(pe_id, block_name))
    a = self.table[pe_id][block_name]
    array = np.array(a)
    return array
    
  def getMean(self, pe_id, block_name):
    array = self.getArray(pe_id, block_name)
    return int(np.mean(array))

  def getVariance(self, pe_id, block_name):
    array = self.getArray(pe_id, block_name)
    return np.var(array)

  def getMeanDict(self, pe_id):
    for k, v in self.table[pe_id].items():
      v = self.getMean(pe_id, k)
      print(k, v)

  def getVarianceDict(self, pe_id):
    for k, v in self.table[pe_id].items():
      v = self.getVariance(pe_id, k)
      print(k, v)

  def getMeanVarianceDict(self, pe_id):
    for k, v in self.table[pe_id].items():
      m = self.getMean(pe_id, k)
      v = self.getVariance(pe_id, k)
      print(k, m, v)

  def getCArray(self, array):
    array_str = "{"
    for cycle in array:
      array_str = array_str + str(cycle) + ","
    array_str = array_str[:-1] + "}"
    return array_str

  def getMinCycle(self):
    min_cycle = int(10000000)
    for pe_id in range(PE_NUM):
      for k, v in self.table[pe_id].items():
        array = self.getArray(pe_id, k)
        constant_cycle = int(np.average(array))
        if constant_cycle > 0 and constant_cycle < min_cycle:
          min_cycle = constant_cycle
    return min_cycle

  def getCTimingCode(self):
    print("#define CORE_NUM " + str(PE_NUM))
    print("#include <string.h>")
    print("#define ISBLOCK(bname) (strcmp(blockname,  #bname ) == 0)")
    print("unsigned long getBlockCycle(unsigned int pe, char* blockname) {")

    # Left 30 for the overhead in case that the critical part is the min.
    overhead = self.getMinCycle() - 30
    print("\tconst unsigned long overhead = " + str(overhead) + ";")

    for pe_id in range(PE_NUM):
      print("\tif (pe == " + str(pe_id) + ") {")
      for k, v in self.table[pe_id].items():
        array = self.getArray(pe_id, k)

        constant_cycle = int(np.average(array))
        ConditionCode = "ISBLOCK(" + k + ")"
        TimingCode = " {return " + str(constant_cycle) + " - overhead;}"
        RetTimingCode = "\t\tif (" + ConditionCode +")" + TimingCode
        print(RetTimingCode)
      print("\t}")
    print("}")


class PatternReader:
  def __init__(self, BlockProfileFile, TimeProfileFile):
    self.f1=open(BlockProfileFile, 'r')
    self.f2=open(TimeProfileFile, 'r')
    assert(self.f1)
    assert(self.f2)
    self.group = LineParserGroup()
    self.EndCode = 1
    self.LiveCode = 0
    self.table = StatisticManager()

  def getID(self, line):
    pe_id = int(line[2])
    assert(pe_id < PE_NUM)
    return pe_id

  def fillBlockLine(self, f):
    line = f.readline()
    if not line:
      #print("block line stops")
      return self.EndCode
    pe_id = self.getID(line)
    self.group.processBlockLine(pe_id, line)
    return self.LiveCode


  def fillTimeLine(self, f):
    line = f.readline()
    if not line:
      #print("time line stops")
      return self.EndCode, 0, "", 0
    pe_id = self.getID(line)
    block_name, time = self.group.processTimeLine(pe_id, line)
    block_name, elapsed_time = self.group.processBlock(pe_id, block_name, time)
    #if block_name == 'raster_hw_e' and pe_id == 1:
      #print(elapsed_time)
    return self.LiveCode, pe_id, block_name, elapsed_time


  def isMinZero(self, emptyflag):
    if emptyflag:
      return 0
    return (self.group.getMinSize() == 0)

  def minOneOrBlockEmpty(self, emptyflag1, emptyflag2):
    if emptyflag2:
      return 0
    if emptyflag1:
      return 1
    return (self.group.getMinSize() > 0)

  def process(self):
    emptyf1 = 0
    emptyf2 = 0
    while (not (emptyf1 and emptyf2)):
      while self.isMinZero(emptyf1):
        emptyf1 = self.fillBlockLine(self.f1)
  
      while(self.minOneOrBlockEmpty(emptyf1, emptyf2)):
        emptyf2, pe_id, block_name, elapsed_time = self.fillTimeLine(self.f2)
        if not emptyf2:
          self.table.add(pe_id, block_name, elapsed_time)

    #for i in range(PE_NUM):
      #print("====================== mean %i ======================" % i)
      #self.table.getMeanVarianceDict(i)

    self.table.getCTimingCode()

filename1= "./block.log"
filename2= "./time.log"
reader = PatternReader(filename1, filename2)
reader.process()
