#include <iostream>
#include <iomanip>
#include <sstream>

#include <systemc>
#include "tlm_utils/simple_target_socket.h"
#include "tlm_utils/simple_initiator_socket.h"

// Caution: The verilator related headers must include after SystemC headers.
//          Otherwise it will finds no SYSTEMC_VERSION and assume you
//          are using oldest SystemC API.
#include "Vsobel.h"

using namespace sc_core;
using namespace tlm;

const char* MSGID = "Sobel-TLM-Wrapper";

#define PRINT_TIME_FORMAT(str) \
      {std::cout << "@ " << std::setw(9) << std::left << sc_time_stamp() \
                 << " " << str << '\n'; }

struct ReturnValue {
  unsigned int iVal1;
  unsigned int iVal2;
};

typedef union word {
  int sint;
  unsigned int uint;
  unsigned char uc[4];
} word;


void print_result_time_format(unsigned int a, unsigned int b, unsigned int c, unsigned int d, 
                              const char* comment_of_number) {
  PRINT_TIME_FORMAT(comment_of_number << ": " << a << " " << b << " " << c << " " << d)
}

// The GCD function used for verification
// unsigned int SoftGCD(unsigned int u, unsigned int v) {
//     while ( v != 0) {
//         unsigned int r = u % v;
//         u = v;
//         v = r;
//     }
//     return u;
// }

ReturnValue SoftSobel(unsigned char m0, unsigned char m1, unsigned char m2, unsigned char m3, unsigned char m5, unsigned char m6, unsigned char m7, unsigned char m8) {
  ReturnValue SWrv;
  unsigned int row_0 = m0 + 2 * m1 + m2;
  unsigned int row_2 = m6 + 2 * m7 + m8;
  unsigned int col_0 = m0 + 2 * m3 + m6;
  unsigned int col_2 = m2 + 2 * m5 + m8;
  if (row_0 > row_2) {
    SWrv.iVal1 = row_0 - row_2;
  }
  else {
    SWrv.iVal1 = row_2 - row_0;
  }
  if (col_0 > col_2) {
    SWrv.iVal2 = col_0 - col_2;
  }
  else {
    SWrv.iVal2 = col_2 - col_0;
  }
  return SWrv;
}

SC_MODULE(SobelTlm)
{
  tlm_utils::simple_target_socket<SobelTlm> targ_skt;

  SC_CTOR(SobelTlm):
    targ_skt("sobel_target_socket"), m_clk_period(sc_time(1, SC_NS))
  {
    // reset phase
    m_sobel.reset = 1;
    // make a rising edge to clock
    step();
    // end of reset phase
    m_sobel.reset = 0;
    // register the TLM transport function
    targ_skt.register_b_transport(this, &SobelTlm::b_transport);
    SC_THREAD(Sobel_process);
  }

  void step() {
    // make a rising edge to clock
    m_sobel.clock = 0;
    m_sobel.eval();
    m_sobel.clock = 1;
    m_sobel.eval();
  }

  void Sobel_process()
  {
    while (true)
    {
      // unsigned int num = num_fifo.read();
      // unsigned int a = num & 0xffff;
      // unsigned int b = num >> 16 & 0xffff;
      word package;
      package.uint = num_fifo.read();
      m_sobel.io_m0 = package.uc[3];
      m_sobel.io_m1 = package.uc[2];
      m_sobel.io_m2 = package.uc[1];
      m_sobel.io_m3 = package.uc[0];

      package.uint = num_fifo.read();
      m_sobel.io_m5 = package.uc[3];
      m_sobel.io_m6 = package.uc[2];
      m_sobel.io_m7 = package.uc[1];
      m_sobel.io_m8 = package.uc[0];
      // print_result_time_format(a, b, "GCD receive numbers");

      unsigned int cycle = 0;

      // write inputs and wait for one cycle to make sure the
      // value is writen to internal registers,
      // m_sobel.io_a = a;
      // m_sobel.io_b = b;
      // m_sobel.io_e = 1;
      // make a rising edge to clock
      step();
      // increment the time
      cycle++;
      // turn off the enable so that the internal registers would
      // not be overwritten by IO registers
      // m_sobel.io_e = 0;

      // Check whether the valid signal rise
      // while(! m_sobel.io_v) {
      // make a rising edge to clock
      // step();
      // increment the time
      // cycle++;
      // }

      wait(cycle * m_clk_period);

      result_fifo.write(m_sobel.io_out_X);
      result_fifo.write(m_sobel.io_out_Y);
    }
  }

  // Parse the payload and read from the FIFO or write to the FIFO
  void b_transport( tlm_generic_payload & trans, sc_time & delay )
  {
    tlm_command cmd = trans.get_command();
    unsigned *       ptr = reinterpret_cast<unsigned *>(trans.get_data_ptr());
    unsigned int     len = trans.get_data_length();

    switch (cmd) {
    case TLM_WRITE_COMMAND:
      // if (len != 4) {
      //   SC_REPORT_ERROR(MSGID,
      //                   "GCD should receive 2 16-bit numbers that store in a integer variable.");
      // }
      num_fifo.write(*ptr);
      wait(1 * m_clk_period);
      break;

    case TLM_READ_COMMAND:
      *ptr = result_fifo.read();
      break;

    default:
      break;
    }

    trans.set_response_status( TLM_OK_RESPONSE );
  }

private:
  Vsobel m_sobel;
  sc_fifo<unsigned int> num_fifo;
  sc_fifo<unsigned int> result_fifo;
  const sc_time m_clk_period;
};

SC_MODULE( Tester )
{
  tlm_utils::simple_initiator_socket<Tester> init_skt;

  SC_CTOR( Tester ):
    init_skt( "tester_init_socket" )
  {
    SC_THREAD(test_process);
  }

  void test_process() {
    // unsigned int a = 64;
    // unsigned int b = 48;
    unsigned char m0, m1, m2, m3, m5, m6, m7, m8;
    unsigned seed;
    srand((unsigned)time(NULL));
    m0 = rand() % 255;
    m1 = rand() % 255;
    m2 = rand() % 255;
    m3 = rand() % 255;
    m5 = rand() % 255;
    m6 = rand() % 255;
    m7 = rand() % 255;
    m8 = rand() % 255;
    send_numbers(m0, m1, m2, m3);
    send_numbers(m5, m6, m7, m8);
    // send_numbers(a, b);
    auto result_X = recv_result();
    auto result_Y = recv_result();

    ReturnValue golden = SoftSobel(m0, m1, m2, m3, m5, m6, m7, m8);
    // auto golden = SoftSobel(m0, m1, m2, m3, m5, m6, m7, m8);

    if ((result_X == golden.iVal1) && (result_Y == golden.iVal2)) {
      std::cout << "Test Passed.\n";
	  std::cout << "result_X: " << result_X << ", and golden.iVal1: " << golden.iVal1 << std::endl;
      std::cout << "result_Y: " << result_Y << ", and golden.iVal2: " << golden.iVal2 << std::endl;
    } else {
      std::cout << "Test Failed!\n";
    }
  }

  // Prepare the write payload and send it to GcdTlm
  void send_numbers(const unsigned char a, const unsigned char b, const unsigned char c, const unsigned char d) {
    tlm::tlm_generic_payload* trans = new tlm::tlm_generic_payload;
    sc_time delay = sc_time( 0, SC_NS );
    // unsigned int numbers = (a << 16) | (b);

    word package;
    package.uc[3] = a;
    package.uc[2] = b;
    package.uc[1] = c;
    package.uc[0] = d;
    unsigned int numbers = package.uint;

    trans->set_command( TLM_WRITE_COMMAND );
    trans->set_data_ptr(reinterpret_cast<unsigned char*>(&numbers));
    trans->set_data_length(4);
    print_result_time_format(a, b, c, d, "Tester sending numbers");
    init_skt->b_transport( *trans, delay );
  }

  // Prepare the read payload and send it to GcdTlm
  unsigned int recv_result() {
    tlm::tlm_generic_payload* trans = new tlm::tlm_generic_payload;
    sc_time delay = sc_time( 0, SC_NS );
    unsigned int result;
    trans->set_command( TLM_READ_COMMAND );
    trans->set_data_ptr(reinterpret_cast<unsigned char*>(&result));
    trans->set_data_length(4);
    PRINT_TIME_FORMAT("Tester reading result.")
    init_skt->b_transport( *trans, delay );
    PRINT_TIME_FORMAT("Tester read result: " << result)
    return result;
  }

};

int sc_main(int argc, char *argv[])
{
  SobelTlm sobel("Sobel");
  Tester tester("Tester");
  tester.init_skt.bind(sobel.targ_skt);
  sc_start();
  return 0;
}
