/*
 * jcmarker.c
 *
 * Copyright (C) 1991-1998, Thomas G. Lane.
 * This file is part of the Independent JPEG Group's software.
 * For conditions of distribution and use, see the accompanying README file.
 *
 * This file contains routines to write JPEG datastream markers.
 */

#include "global.h"

typedef enum {			/* JPEG marker codes */
  M_SOF0  = 0xc0,
  M_SOF1  = 0xc1,
  M_SOF2  = 0xc2,
  M_SOF3  = 0xc3,
  
  M_SOF5  = 0xc5,
  M_SOF6  = 0xc6,
  M_SOF7  = 0xc7,
  
  M_JPG   = 0xc8,
  M_SOF9  = 0xc9,
  M_SOF10 = 0xca,
  M_SOF11 = 0xcb,
  
  M_SOF13 = 0xcd,
  M_SOF14 = 0xce,
  M_SOF15 = 0xcf,
  
  M_DHT   = 0xc4,
  
  M_DAC   = 0xcc,
  
  M_RST0  = 0xd0,
  M_RST1  = 0xd1,
  M_RST2  = 0xd2,
  M_RST3  = 0xd3,
  M_RST4  = 0xd4,
  M_RST5  = 0xd5,
  M_RST6  = 0xd6,
  M_RST7  = 0xd7,
  
  M_SOI   = 0xd8,
  M_EOI   = 0xd9,
  M_SOS   = 0xda,
  M_DQT   = 0xdb,
  M_DNL   = 0xdc,
  M_DRI   = 0xdd,
  M_DHP   = 0xde,
  M_EXP   = 0xdf,
  
  M_APP0  = 0xe0,
  M_APP1  = 0xe1,
  M_APP2  = 0xe2,
  M_APP3  = 0xe3,
  M_APP4  = 0xe4,
  M_APP5  = 0xe5,
  M_APP6  = 0xe6,
  M_APP7  = 0xe7,
  M_APP8  = 0xe8,
  M_APP9  = 0xe9,
  M_APP10 = 0xea,
  M_APP11 = 0xeb,
  M_APP12 = 0xec,
  M_APP13 = 0xed,
  M_APP14 = 0xee,
  M_APP15 = 0xef,
  
  M_JPG0  = 0xf0,
  M_JPG13 = 0xfd,
  M_COM   = 0xfe,
  
  M_TEM   = 0x01,
  
  M_ERROR = 0x100
} JPEG_MARKER;

/*
 * Basic output routines.
 *
 * Note that we do not support suspension while writing a marker.
 * Therefore, an application using suspension must ensure that there is
 * enough buffer space for the initial markers (typ. 600-700 bytes) before
 * calling jpeg_start_compress, and enough space to write the trailing EOI
 * (a few bytes) before calling jpeg_finish_compress.  Multipass compression
 * modes are not supported at all with suspension, so those two are the only
 * points where markers will be written.
 */

void
emit_byte (int val)
/* Emit a byte */
{
    fputc(val, jpg_fp);
}

void
emit_marker (JPEG_MARKER mark)
/* Emit a marker code */
{
  emit_byte(0xFF);
  emit_byte(mark);
}

void
emit_2bytes (int value)
/* Emit a 2-byte integer; these are always MSB first in JPEG files */
{
  emit_byte((value >> 8) & 0xFF);
  emit_byte(value & 0xFF);
}

/*
 * Routines to write specific marker types.
 */

int
emit_dqt (int index, int *qtbl)
/* Emit a DQT marker */
/* Returns the precision used (0 = 8bits, 1 = 16bits) for baseline checking */
{
  int prec;
  int i;

  if (qtbl == NULL) {
    fprintf(stderr, "%s: no quantization table\n", "cjpeg");
    exit(1);
  }

  prec = 0;
  for (i = 0; i < DCTSIZE2; i++) {
    if (qtbl[i] > 255)
      prec = 1;
  }

  emit_marker(M_DQT);

  emit_2bytes(prec ? DCTSIZE2*2 + 1 + 2 : DCTSIZE2 + 1 + 2);

  emit_byte(index + (prec<<4));

  for (i = 0; i < DCTSIZE2; i++) {
    /* The table entries must be emitted in zigzag order. */
    unsigned int qval = qtbl[jpeg_natural_order[i]];
    if (prec)
      emit_byte((int) (qval >> 8));
    emit_byte((int) (qval & 0xFF));
  }

  return prec;
}


void
emit_dht (int index, int is_ac, unsigned char bits[], unsigned char huffval[])
/* Emit a DHT marker */
{
  int length, i;
  
  if (is_ac) {
    index += 0x10;		/* output index has AC bit set */
  }

  if (bits == NULL || huffval == NULL) {
    fprintf(stderr, "%s: no Huffman table\n", "cjpeg");
    exit(1);
  }
  
  emit_marker(M_DHT);
    
  length = 0;
  for (i = 1; i <= 16; i++)
    length += bits[i];
    
  emit_2bytes(length + 2 + 1 + 16);
  emit_byte(index);
    
  for (i = 1; i <= 16; i++)
    emit_byte(bits[i]);
    
  for (i = 0; i < length; i++)
    emit_byte(huffval[i]);
    
}

void
emit_sof (JPEG_MARKER code)
/* Emit a SOF marker */
{
  emit_marker(code);
  
  emit_2bytes(3 * 3 + 2 + 5 + 1); /* length */

  /* Make sure image isn't bigger than SOF field can handle */
  if ((long) org_h > 65535L || (long) org_w > 65535L) {
    fprintf(stderr, "%s: image too big\n", "cjpeg");
    exit(1);
  }

  emit_byte(8);                     /* data precision */
  emit_2bytes(org_h);               /* image height */
  emit_2bytes(org_w);               /* image width */

  emit_byte(3);                     /* # of components */

  emit_byte(1);                     /* component ID for Y */
  emit_byte((2 << 4) + 2);          /* h samp. = 2, v samp. = 2 */
  emit_byte(0);                     /* q table no. */
  emit_byte(2);                     /* component ID for Cb */
  emit_byte((1 << 4) + 1);          /* h samp. = 1, v samp. = 1 */
  emit_byte(1);                     /* q table no. */
  emit_byte(3);                     /* component ID for Cr */
  emit_byte((1 << 4) + 1);          /* h samp. = 1, v samp. = 1 */
  emit_byte(1);                     /* q table no. */
}


void
emit_sos ()
/* Emit a SOS marker */
{
  emit_marker(M_SOS);
  
  emit_2bytes(2 * 3 + 2 + 1 + 3); /* length */
  
  emit_byte(3);
  
  emit_byte(1);                   /* component ID for Y */
  emit_byte((0 << 4) + 0);        /* DC table no. = 0, AC table no. = 0 */
  emit_byte(2);                   /* component ID for Cb */
  emit_byte((1 << 4) + 1);        /* DC table no. = 1, AC table no. = 1 */
  emit_byte(3);                   /* component ID for Cr */
  emit_byte((1 << 4) + 1);        /* DC table no. = 1, AC table no. = 1 */

  emit_byte(0);                   /* Ss */
  emit_byte(63);                  /* Se */
  emit_byte((0 << 4) + 0);        /* Ah, Al */
}

void
emit_jfif_app0 ()
/* Emit a JFIF-compliant APP0 marker */
{
  /*
   * Length of APP0 block	(2 bytes)
   * Block ID			(4 bytes - ASCII "JFIF")
   * Zero byte			(1 byte to terminate the ID string)
   * Version Major, Minor	(2 bytes - major first)
   * Units			(1 byte - 0x00 = none, 0x01 = inch, 0x02 = cm)
   * Xdpu			(2 bytes - dots per unit horizontal)
   * Ydpu			(2 bytes - dots per unit vertical)
   * Thumbnail X size		(1 byte)
   * Thumbnail Y size		(1 byte)
   */
  
  emit_marker(M_APP0);
  
  emit_2bytes(2 + 4 + 1 + 2 + 1 + 2 + 2 + 1 + 1); /* length */

  emit_byte(0x4A);	/* Identifier: ASCII "JFIF" */
  emit_byte(0x46);
  emit_byte(0x49);
  emit_byte(0x46);
  emit_byte(0);
  emit_byte(1); /* Version fields */
  emit_byte(1);
  emit_byte(0); /* Pixel size information */
  emit_2bytes(1);
  emit_2bytes(1);
  emit_byte(0);		/* No thumbnail image */
  emit_byte(0);
}


/*
 * Write datastream header.
 * This consists of an SOI and optional APPn markers.
 * We recommend use of the JFIF marker, but not the Adobe marker,
 * when using YCbCr or grayscale data.  The JFIF marker should NOT
 * be used for any other JPEG colorspace.  The Adobe marker is helpful
 * to distinguish RGB, CMYK, and YCCK colorspaces.
 * Note that an application can write additional header markers after
 * jpeg_start_compress returns.
 */

void
write_file_header ()
{
  emit_marker(M_SOI);	/* first the SOI */

  emit_jfif_app0();
}


/*
 * Write frame header.
 * This consists of DQT and SOFn markers.
 * Note that we do not emit the SOF until we have emitted the DQT(s).
 * This avoids compatibility problems with incorrect implementations that
 * try to error-check the quant table numbers as soon as they see the SOF.
 */

void
write_frame_header ()
{
  int prec;
  
  /* Emit DQT for each quantization table.
   * Note that emit_dqt() suppresses any duplicate tables.
   */
  prec = 0;
  prec += emit_dqt(0, qtbl0);
  prec += emit_dqt(1, qtbl1);
  /* now prec is nonzero iff there are any 16-bit quant tables. */

  emit_sof(M_SOF0);	/* SOF code for baseline implementation */
}


/*
 * Write scan header.
 * This consists of DHT or DAC markers, optional DRI, and SOS.
 * Compressed data will be written following the SOS.
 */

void
write_scan_header ()
{
  /* Emit Huffman tables.
   * Note that emit_dht() suppresses any duplicate tables.
   */
  /* Sequential mode: need both DC and AC tables */
  emit_dht(0, FALSE, dc_bits0, dc_huffval0);
  emit_dht(0, TRUE, ac_bits0, ac_huffval0);
  emit_dht(1, FALSE, dc_bits1, dc_huffval1);
  emit_dht(1, TRUE, ac_bits1, ac_huffval1);

  emit_sos();
}


/*
 * Write datastream trailer.
 */

void
write_file_trailer ()
{
  emit_marker(M_EOI);
}
