#include <string>
#include "tlm.h"
#include "ovpworld.org/modelSupport/tlmPlatform/1.0/tlm2.0/tlmPlatform.hpp"
#include "ovpworld.org/modelSupport/tlmDecoder/1.0/tlm2.0/tlmDecoder.hpp"
#include "ovpworld.org/memory/ram/1.0/tlm2.0/tlmMemory.hpp"
#include "riscv.ovpworld.org/processor/riscv/1.0/tlm2.0/processor.igen.hpp"


static const unsigned int stack_mem_size = 0x100000u;
static const unsigned int stack_mem_start = 0x100000000u - stack_mem_size;
static const unsigned int stack_mem_end = 0x100000000u - 1;

static const unsigned int share_mem_size = 0x80000000u;
static const unsigned int share_mem_start = 0; 
static const unsigned int share_mem_end = share_mem_size - 1;

std::string concat(std::string s, int n) {
  return s + std::to_string(n);
}


#define DMI_CODE true
////////////////////////////////////////////////////////////////////////////////
//                                MulticoreOVP Class //
////////////////////////////////////////////////////////////////////////////////
template <int NCORE>
struct MulticoreOVP : sc_module {
  icmTLMPlatform Platform;

  riscv* cpu[NCORE];
  ram* stack_mem[NCORE];
  decoder<2, 2>* tile_bus[NCORE];

  decoder<NCORE, 1>* sys_bus;

  ram* shared_mem;


  MulticoreOVP(const sc_core::sc_module_name name)
      : sc_module(name),
        Platform("icm", ICM_VERBOSE | ICM_STOP_ON_CTRLC |
                            ICM_ENABLE_IMPERAS_INTERCEPTS) {
    init_components();
    connect_components();
  }

  ~MulticoreOVP() {
    for (int i = 0; i < NCORE; ++i) {
      delete cpu[i];
      delete tile_bus[i];
      delete stack_mem[i];
    }
    delete sys_bus;
    delete shared_mem;
  }

  void load_program(const char* program) {
    for (int i = 0; i < NCORE; ++i) {
      cpu[i]->loadLocalMemory(
          program, (icmLoaderAttrs)(ICM_LOAD_VERBOSE | ICM_SET_START));
    }
  }


  void assert_reset() {
    for (int i = 0; i < NCORE; ++i) {
      cpu[i]->reset.write(1);
    }
  }

  void deassert_reset() {
    for (int i = 0; i < NCORE; ++i) {
      cpu[i]->reset.write(0);
    }
  }
  

 private:
  void init_components() {
    for (int i = 0; i < NCORE; ++i) {
      cpu[i] = new riscv
        (concat("cpu", i).c_str(), i, ICM_ATTR_DEFAULT, NULL,
         "/users/others/jjlab/lib/Imperas-semihosing/manycore-semihosing.so"
         , 32, DMI_CODE, 0);

      tile_bus[i] = new decoder<2, 2>
        (concat("tile_bus_", i).c_str());

      stack_mem[i] = new ram
        (concat("stack_mem", i).c_str(), "sp1", stack_mem_size - 1);
    }
    sys_bus = new decoder<NCORE, 1> ("sys_bus");
    shared_mem = new ram ("shared_mem", "sp1", share_mem_size - 1);
  }

  void connect_components() {
    for (int i = 0; i < NCORE; ++i) {
      // connect cpu and tile bus
      cpu[i]->INSTRUCTION.socket(tile_bus[i]->target_socket[0]);
      cpu[i]->DATA.socket(tile_bus[i]->target_socket[1]);

      // connect stack memory and tile bus
      tile_bus[i]->initiator_socket[0](stack_mem[i]->sp1);
      tile_bus[i]->setDecode(0, stack_mem_start, stack_mem_end);

      // connect tile bus and system bus
      tile_bus[i]->initiator_socket[1](sys_bus->target_socket[i]);
      tile_bus[i]->setDecode(1, share_mem_start, share_mem_end);
    }

    // connect system bus and shared memory
    sys_bus->initiator_socket[0](shared_mem->sp1); 
    sys_bus->setDecode(0, share_mem_start, share_mem_end);
  }

}; /* MulticoreOVP */

int sc_main(int argc, char* argv[]) {
  const char* exe;
  if (argc == 2) {
    exe = argv[1];
  } else {
    cout << "Usage: multicoreOVP <application>" << endl;
    cout << "       Please specify application" << endl;
    return 1;
  }

  const int NCORE = 4;
  MulticoreOVP<NCORE> multicoreOVP("multicoreOVP");

  cout << "Using SystemC TLM2.0 Memory with DMI" << endl;
  multicoreOVP.load_program(exe);


  cout << "Starting sc_main." << endl;
  sc_core::sc_start();

  return 0;  // return okay status
}

